"""Rich Printer."""

import logging
import shutil
from datetime import datetime

from beartype import beartype
from beartype.typing import Any, List, Optional
from rich.console import Console
from rich.text import Text

from .styles import Styles, get_name


@beartype
def rich_printer(  # noqa: CAC001
    message: str,
    *,
    is_header: bool,
    _this_level: int,
    _is_text: bool,
    # Logger-specific parameters that need to be initialized with partial(...)
    _console: Console,
    _styles: Styles,
    _keys_on_own_line: Optional[List[str]] = None,
    **kwargs: Any,
) -> None:
    """Generic log writer.."""
    text = Text()
    if _is_text:
        if is_header:
            print()  # noqa: T201
        text.append(f'{message}', style=_styles.message)
    else:
        timestamp = kwargs.pop('timestamp', datetime.now())  # noqa: DTZ005
        text.append(f'{timestamp: <28} ', style=_styles.timestamp)
        text.append('[', style=_styles.timestamp)
        level_style = _styles.get_style(level=_this_level)
        text.append(f'{get_name(level=_this_level): <7}', style=level_style)
        text.append(']', style=_styles.timestamp)
        text.append(f' {message}', style=_styles.message)

    full_lines = []
    for key in _keys_on_own_line or []:
        if line := kwargs.pop(key, None):
            full_lines.append((key, line))  # noqa: PERF401
    for key, value in kwargs.items():
        text.append(f' {key}=', style=_styles.key)
        text.append(f'{value!s}', style=_styles.value)
    _console.print(text)
    for key, line in full_lines:
        new_text = Text()
        new_text.append(f' ∟ {key}', style=_styles.key)
        new_text.append(f': {line}', style=_styles.value_own_line)
        _console.print(new_text)

    if _this_level == logging.CRITICAL:
        term_width, _height = shutil.get_terminal_size((100, 50))
        _console.print_exception(
            extra_lines=1,
            max_frames=6,
            show_locals=False,
            width=term_width,
        )
        # > or 'from rich.traceback import install; install(show_locals=True)'



"""Track delegated processes with rich progress meters.

Based on: https://www.deanmontgomery.com/2022/03/24/rich-progress-and-multiprocessing

"""

import math
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from multiprocessing.managers import DictProxy
from time import sleep

from beartype import beartype
from beartype.typing import Any, Callable, List, TypeVar, Union
from rich.progress import BarColumn, Progress, ProgressColumn, TaskID, TimeElapsedColumn, TimeRemainingColumn

_ItemT = TypeVar('_ItemT', bound=Any)
"""Iterated item in the data."""


_DelegatedTask = Callable[
    [
        TaskID,
        DictProxy,  # type: ignore[type-arg]
        list[_ItemT],
    ],
    Any,
]


@beartype
def _chunked(data: List[_ItemT], count: int) -> List[List[_ItemT]]:
    """Chunk the list of data into equally sized lists."""
    # TODO: See below link for other options for chunking
    #   https://realpython.com/how-to-split-a-python-list-into-chunks/
    size = len(data)
    chunk_size, chunk_rem = size // count, size % count
    chunk_size += int(math.ceil(chunk_rem / size))
    return [data[ix : ix + chunk_size] for ix in range(0, size, chunk_size)]


@beartype
def pretty_process(
    delegated_task: _DelegatedTask,  # type: ignore[type-arg]
    *,
    data: List[_ItemT],
    num_workers: int = 3,
    num_cpus: int = 4,
) -> Any:
    """Run a task in parallel to process all provided data.

    Uses `rich` to display pretty progress bars

    Args:
        delegated_task: must call `shared_progress[task_id] += 1` on each item in data
        data: the list of data to distribute
        num_workers: number of worker processes
        num_cpus: number of CPUs

    """
    # Docs: https://rich.readthedocs.io/en/latest/progress.html
    columns: List[Union[str, ProgressColumn]] = [
        '[progress.description]{task.description}',
        BarColumn(),
        '[progress.percentage]{task.percentage:>3.0f}%',
        TimeRemainingColumn(),
        TimeElapsedColumn(),
    ]
    with Progress(*columns, refresh_per_second=1) as progress:  # noqa: SIM117 (Py>3.9)
        # Share state between process and workers
        with multiprocessing.Manager() as manager:
            shared_progress = manager.dict()
            jobs = []
            totals = {}
            task_id_all = progress.add_task('[green]All jobs progress:')

            with ProcessPoolExecutor(max_workers=num_workers) as executor:
                for ix, chunk in enumerate(_chunked(data, count=num_cpus)):
                    task_id = progress.add_task(f'task {ix}')
                    shared_progress[task_id] = 0
                    totals[task_id] = len(chunk)
                    jobs.append(executor.submit(delegated_task, task_id, shared_progress, chunk))

                # Update progress bar from shared state
                remaining = len(jobs)
                while remaining:
                    n_done = 0
                    for task_id, latest in shared_progress.items():
                        n_done += latest
                        progress.update(task_id, completed=latest, total=totals[task_id])
                    progress.update(task_id_all, completed=n_done, total=len(data))
                    remaining = len(jobs) - sum(job.done() for job in jobs)

                # Collect results and catch and errors
                return [job.result() for job in jobs]


# Note: can't use beartype on a delegated_task & this function can't be in the if-block below
def __long_task(task_id: int, shared_progress: DictProxy, data: List[_ItemT]) -> Any:  # type: ignore[type-arg]
    """Example long task."""
    for _val in data:
        sleep(1)  # nosemgrep: python.lang.best-practice.sleep.arbitrary-sleep
        shared_progress[task_id] += 1
    return True


if __name__ == '__main__':

    def _demo() -> None:
        """Run demo with: 'poetry run python -m shoal.pretty_process'."""
        # Resolve number of cores or specified maximum
        num_cpus = 4
        try:
            import psutil  # pyright: ignore[reportMissingModuleSource] # noqa: PLC0415

            num_cpus = psutil.cpu_count(logical=False)
        except Exception as exc:
            print(exc)  # noqa: T201

        result = pretty_process(__long_task, data=[*range(23)], num_workers=num_cpus)
        print(result)  # noqa: T201

    _demo()



"""Run shell commands."""

import asyncio
import subprocess  # nosec # noqa: S404
import sys
from io import BufferedReader, StringIO, TextIOWrapper
from pathlib import Path
from time import time

from beartype import beartype
from beartype.typing import Callable, Optional, Union

from .log import LOGGER


@beartype
def capture_shell(
    cmd: str, *, timeout: int = 120, cwd: Optional[Path] = None, printer: Optional[Callable[[str], None]] = None,
) -> str:
    """Run shell command, return the output, and optionally print in real time.

    Inspired by: https://stackoverflow.com/a/38745040/3219667

    Args:
        cmd: shell command
        timeout: process timeout. Defaults to 2 minutes
        cwd: optional path for shell execution
        printer: optional callable to output the lines in real time

    Returns:
        str: stripped output

    Raises:
        CalledProcessError: if return code is non-zero

    """
    LOGGER.debug('Running', cmd=cmd, timeout=timeout, cwd=cwd, printer=printer)

    start = time()
    lines = []
    with subprocess.Popen(  # nosec  # nosemgrep
        cmd, cwd=cwd,
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True,
        shell=True,  # noqa: S602
    ) as proc:
        stdout: Union[BufferedReader, StringIO, TextIOWrapper] = proc.stdout  # type: ignore[assignment]
        return_code = None
        while return_code is None:
            if timeout != 0 and time() - start >= timeout:
                proc.kill()
                break
            if line := stdout.readline():
                lines.append(line)
                if printer:
                    printer(line.rstrip())  # type: ignore[arg-type]
            else:
                return_code = proc.poll()

    output = ''.join(lines)  # type: ignore[arg-type]
    if return_code != 0:
        raise subprocess.CalledProcessError(returncode=return_code or 404, cmd=cmd, output=output)
    return output


async def _capture_shell_async(cmd: str, *, cwd: Optional[Path] = None) -> str:
    proc = await asyncio.create_subprocess_shell(  # nosec  # nosemgrep
        cmd, cwd=cwd,
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
        shell=True,  # noqa: S604
    )

    stdout, _stderr = await proc.communicate()
    output = stdout.decode().strip()
    if proc.returncode != 0:
        raise subprocess.CalledProcessError(returncode=proc.returncode or 404, cmd=cmd, output=output)
    return output


async def capture_shell_async(cmd: str, *, timeout: int = 120, cwd: Optional[Path] = None) -> str:
    """Run a shell command asynchronously and return the output.

    ```py
    print(asyncio.run(capture_shell_async('ls ~/.config')))
    ```

    Args:
        cmd: shell command
        timeout: process timeout. Defaults to 2 minutes
        cwd: optional path for shell execution

    Returns:
        str: stripped output

    Raises:
        CalledProcessError: if return code is non-zero

    """
    LOGGER.debug('Running', cmd=cmd, timeout=timeout, cwd=cwd)
    return await asyncio.wait_for(_capture_shell_async(cmd=cmd, cwd=cwd), timeout=timeout)


@beartype
def run_shell(cmd: str, *, timeout: int = 120, cwd: Optional[Path] = None) -> None:
    """Run a shell command without capturing the output.

    Args:
        cmd: shell command
        timeout: process timeout. Defaults to 2 minutes
        cwd: optional path for shell execution

    Raises:
        CalledProcessError: if return code is non-zero

    """
    LOGGER.debug('Running', cmd=cmd, timeout=timeout, cwd=cwd)

    subprocess.run(  # nosemgrep
        cmd, timeout=timeout or None, cwd=cwd,
        stdout=sys.stdout, stderr=sys.stderr, check=True,
        shell=True,  # noqa: S602,  # nosec
    )



"""Structlog Logger."""

import logging

import structlog
from beartype import beartype
from beartype.typing import Any


@beartype
def structlog_logger(
    message: str,
    *,
    is_header: bool,
    _this_level: int,
    _is_text: bool,
    # Logger-specific parameters that need to be initialized with partial(...)
    **kwargs: Any,
) -> None:
    logger = structlog.get_logger()
    log = {
        logging.CRITICAL: logger.exception,
        logging.ERROR: logger.error,
        logging.WARNING: logger.warning,
        logging.INFO: logger.info,
        logging.DEBUG: logger.debug,
        logging.NOTSET: logger.debug,
    }.get(_this_level, logger.msg)
    log(message, is_header=is_header, _this_level=_this_level, _is_text=_is_text, **kwargs)



"""Styles."""

import logging
from dataclasses import dataclass

from beartype import beartype
from beartype.typing import Optional


@dataclass
class Colors:
    """Based on Tokyo Night: https://github.com/folke/tokyonight.nvim#-extras."""

    level_error: str = '#e77d8f'
    level_warn: str = '#d8b172'
    level_info: str = '#a8cd76'
    level_debug: str = '#82a1f1'
    level_fallback: str = '#b69bf1'


@dataclass
class Styles:
    """Inspired by `loguru` and `structlog` and used in `tail-jsonl`.

    https://rich.readthedocs.io/en/latest/style.html

    Inspired by: https://github.com/Delgan/loguru/blob/07f94f3c8373733119f85aa8b9ca05ace3325a4b/loguru/_defaults.py#L31-L73

    And: https://github.com/hynek/structlog/blob/bcfc7f9e60640c150bffbdaeed6328e582f93d1e/src/structlog/dev.py#L126-L141

    """

    timestamp: str = '#8DAAA1'
    message: str = 'bold'

    colors: Optional[Colors] = None

    # triadic from: https://coolors.co/a28eab
    key: str = '#8DAAA1'
    value: str = '#A28EAB'
    value_own_line: str = '#AAA18D'

    def get_style(self, *, level: int) -> str:
        """Return the right style for the specified level."""
        if not self.colors:
            self.colors = Colors()
        return {
            logging.CRITICAL: self.colors.level_error,
            logging.ERROR: self.colors.level_error,
            logging.WARNING: self.colors.level_warn,
            logging.INFO: self.colors.level_info,
            logging.DEBUG: self.colors.level_debug,
        }.get(level, self.colors.level_fallback)


@beartype
def get_level(*, name: str) -> int:
    """Return the logging level based on the provided name."""
    return {
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'WARN': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG,
    }.get(name.upper(), logging.NOTSET)


@beartype
def get_name(*, level: int) -> str:
    """Return the logging name based on the provided level.

    https://docs.python.org/3.11/library/logging.html#logging-levels

    """
    return {
        logging.CRITICAL: 'EXCEPTION',
        logging.ERROR: 'ERROR',
        logging.WARNING: 'WARNING',
        logging.INFO: 'INFO',
        logging.DEBUG: 'DEBUG',
        logging.NOTSET: 'NOTSET',
    }.get(level, '')


STYLES = Styles()



"""Export tomllib."""

try:
    import tomllib  # type: ignore[import-not-found] # pyright: ignore[reportMissingImports]
except ModuleNotFoundError:  # pragma: no cover
    import tomli as tomllib  # type: ignore[import-not-found] # pyright: ignore[reportMissingImports]
__all__ = ('tomllib',)
